@@ -2,7 +2,7 @@ import json
 import os
 from base64 import b64encode
 from datetime import datetime, timedelta
-from urllib.parse import urljoin
+from urllib.parse import quote_plus, urljoin
 
 import requests
 
@@ -10,6 +10,7 @@ SYPHT_API_BASE_ENDPOINT = "https://api.sypht.com"
 SYPHT_AUTH_ENDPOINT = "https://auth.sypht.com/oauth2/token"
 SYPHT_LEGACY_AUTH_ENDPOINT = "https://login.sypht.com/oauth/token"
 SYPHT_OAUTH_COMPANY_ID_CLAIM_KEY = "https://api.sypht.com/companyId"
+TOKEN_EXPIRY_BUFFER_SECONDS = 10
 
 
 def _iter_chunked_sequence(seq, size):
@@ -17,7 +18,7 @@ def _iter_chunked_sequence(seq, size):
         yield seq[pos : pos + size]
 
 
-class SyphtClient(object):
+class SyphtClient:
     API_ENV_KEY = "SYPHT_API_KEY"
 
     def __init__(
@@ -49,13 +50,15 @@ class SyphtClient(object):
             if not env_key:
                 raise ValueError(
                     "Missing API key configuration. Add "
-                    + self.API_ENV_KEY
-                    + " to the environment or "
+                    + f"{self.API_ENV_KEY}='<client_id>:<client_secret>' to the environment or "
                     + "directly pass client_id and client_secret parameters to the client constructor."
                 )
             elif env_key and len(key_parts) != 2:
                 raise ValueError(
-                    "Invalid " + self.API_ENV_KEY + " environment variable configured"
+                    f"Invalid {self.API_ENV_KEY} environment variable configured. "
+                    + "<client_id> and <client_secret> must be provided as a single, "
+                    + "colon-separated environment variable, i.e: "
+                    + f"export {self.API_ENV_KEY}='<client_id>:<client_secret>'"
                 )
             client_id, client_secret = key_parts
 
@@ -130,16 +133,24 @@ class SyphtClient(object):
     def _authenticate_client(self):
         if "/oauth/" in self.auth_endpoint:
             access_token, expires_in = self._authenticate_v1(
-                self.auth_endpoint, self.client_id, self._client_secret, audience=self.audience
+                self.auth_endpoint,
+                self.client_id,
+                self._client_secret,
+                audience=self.audience,
             )
         elif "/oauth2/" in self.auth_endpoint:
             access_token, expires_in = self._authenticate_v2(
-                self.auth_endpoint, self.client_id, self._client_secret, audience=self.audience
+                self.auth_endpoint,
+                self.client_id,
+                self._client_secret,
+                audience=self.audience,
             )
         else:
-            raise ValueError(f"Invalid authentication endpoint: {auth_endpoint}")
+            raise ValueError(f"Invalid authentication endpoint: {self.auth_endpoint}")
 
-        self._auth_expiry = datetime.utcnow() + timedelta(seconds=expires_in)
+        self._auth_expiry = datetime.utcnow() + timedelta(
+            seconds=expires_in - TOKEN_EXPIRY_BUFFER_SECONDS
+        )
         self._access_token = access_token
 
     def _get_headers(self, **headers):
@@ -452,6 +463,8 @@ class SyphtClient(object):
 
     def get_entity(self, entity_id, entity_type, company_id=None, endpoint=None):
         company_id = company_id or self.company_id
+        entity_id = quote_plus(entity_id)
+        entity_type = quote_plus(entity_type)
         endpoint = urljoin(
             endpoint or self.base_endpoint,
             f"storage/{company_id}/entity/{entity_type}/{entity_id}",
@@ -461,8 +474,38 @@ class SyphtClient(object):
         headers["Content-Type"] = "application/json"
         return self._parse_response(self.requests.get(endpoint, headers=headers))
 
-    def list_entities(self, entity_type, company_id=None, page=None, limit=None, endpoint=None):
+    def get_many_entities(self, entity_type, entities, company_id=None, endpoint=None):
+        """
+        Get entities in bulk.
+
+        Entities should be list with structure:
+        [
+            {"entity_id": "id_0"},
+            {"entity_id": "id_1"},
+            {"entity_id": "id_2"},
+            ...
+        ]
+        """
+        if entities is None or not isinstance(entities, list):
+            raise ValueError("Expected a list of entities")
+
         company_id = company_id or self.company_id
+        entity_type = quote_plus(entity_type)
+        endpoint = urljoin(
+            endpoint or self.base_endpoint,
+            f"storage/{company_id}/entitysearch/{entity_type}/by_id",
+        )
+        headers = self._get_headers()
+        headers["Accept"] = "application/json"
+        headers["Content-Type"] = "application/json"
+        return self._parse_response(
+            self.requests.post(endpoint, data=json.dumps(entities), headers=headers)
+        )
+
+    def list_entities(self, entity_type, company_id=None, page=None, limit=None, endpoint=None):
+        """Get list of entity_ids by pagination."""
+        company_id = company_id or self.company_id
+        entity_type = quote_plus(entity_type)
         endpoint = urljoin(
             endpoint or self.base_endpoint,
             f"storage/{company_id}/entitysearch/{entity_type}",
@@ -477,8 +520,34 @@ class SyphtClient(object):
             params["limit"] = int(limit)
         return self._parse_response(self.requests.get(endpoint, headers=headers, params=params))
 
+    def get_all_entity_ids(self, entity_type, verbose=True, company_id=None, endpoint=None):
+        """Get all entity_ids for specified entity_type.
+
+        Returns list of objects if verbose (by default):
+        [{"entity_id": "id_0"}, {"entity_id": "id_1"}, ...]
+
+        Returns list of entity_id if not verbose:
+        ["id_0", "id_1", ...]
+        """
+        entity_ids = []
+        next_page = True
+        while next_page:
+            if next_page is True:
+                next_page = None  # first page request
+            res = self.list_entities(entity_type, company_id, page=next_page)
+            next_page = res.get("next_page")
+            if verbose:
+                entity_ids.extend(
+                    [{"entity_id": entity_id} for entity_id in res.get("entities")]
+                )
+            else:
+                entity_ids.extend(res.get("entities"))
+        return entity_ids
+
     def set_entity(self, entity_id, entity_type, data, company_id=None, endpoint=None):
         company_id = company_id or self.company_id
+        entity_id = quote_plus(entity_id)
+        entity_type = quote_plus(entity_type)
         endpoint = urljoin(
             endpoint or self.base_endpoint,
             f"storage/{company_id}/entity/{entity_type}/{entity_id}",
@@ -492,6 +561,8 @@ class SyphtClient(object):
 
     def delete_entity(self, entity_id, entity_type, company_id=None, endpoint=None):
         company_id = company_id or self.company_id
+        entity_id = quote_plus(entity_id)
+        entity_type = quote_plus(entity_type)
         endpoint = urljoin(
             endpoint or self.base_endpoint,
             f"storage/{company_id}/entity/{entity_type}/{entity_id}",
@@ -519,6 +590,7 @@ class SyphtClient(object):
             raise ValueError("Expected a list of entities")
 
         company_id = company_id or self.company_id
+        entity_type = quote_plus(entity_type)
         endpoint = urljoin(
             endpoint or self.base_endpoint,
             f"storage/{company_id}/bulkentity/{entity_type}/",
@@ -542,8 +614,10 @@ class SyphtClient(object):
         exact = exact or {}
         fuzzy = fuzzy or {}
         company_id = company_id or self.company_id
+        entity_type = quote_plus(entity_type)
         endpoint = urljoin(
-            endpoint or self.base_endpoint, f"storage/{company_id}/entitysearch/{entity_type}/"
+            endpoint or self.base_endpoint,
+            f"storage/{company_id}/entitysearch/{entity_type}/",
         )
         headers = self._get_headers()
         headers["Accept"] = "application/json"
